/*********************************************************************
 This code is built on Adafruits nRF52 arduino libraries
 https://learn.adafruit.com/wireless-untztrument-using-ble-midi/overview

 Adafruit invests time and resources on developing open source code,
 please support Adafruit and open-source hardware by purchasing
 products from Adafruit!
*********************************************************************/

#include <bluefruit.h>
#include <MIDI.h>
#include <Wire.h>
#include <U8g2lib.h>
#include <SPI.h>
#include <RotaryEncoder.h>


// Rotary encoder pin definitions
#define ROTARY_ENC_PINA   12
#define ROTARY_ENC_PINB   13

// Foot-switch pin definitions
#define FS_1_A_PIN    7
#define FS_1_B_PIN    10
#define FS_2_A_PIN    9
#define FS_2_B_PIN    11

// Pedal pin definitions
#define PEDAL1_RING_PIN   A3
#define PEDAL1_TIP_PIN    A1
#define PEDAL2_RING_PIN   A2
#define PEDAL2_TIP_PIN    A0

// Pedal status definitions
#define PEDAL_NOT_CONNECTED 0     // No expression pedal connected
#define PEDAL_CONNECTED     1     // Expression pedal connected
#define PEDAL_ERROR         2     // Wrong pedal connected (Ring connection pulled to ground, typically means either mono JACK or FootSwitch connected)

// Constant definitions
#define DEBOUNCE            50
#define ADC_RESOLUTION      10    // 10 bit
#define ADC_MAX_VALUE       1024  // Max value from ADC at 10bit resolution
#define ADC_RESULT_DIVIDER  8     // Division factor to create midi (0-127) from ADC result

typedef struct                      // Struct for holding each controllers data
{
  byte cccode;
  byte value;
  byte previousSentValue;
  byte channel;
} MIDIControlChangeData;

byte midichannel = 1;               // Midi channel to transmit on

MIDIControlChangeData MidiData[37] =   // define an array holding the cc codes, value, previous value sent over midi and channel for all pots, buttons and pedals
{//{CC code, value, previous sent value, midichannel}
  // pots
  {0x1A,0,0,midichannel},  // MidiData[0] = Pot[0]  : CC 0x1A = Undefined
  {0x1B,0,0,midichannel},  // MidiData[1] = Pot[1]  : CC 0x1B = Undefined
  {0x1C,0,0,midichannel},  // MidiData[2] = Pot[2]  : CC 0x1C = Undefined
  {0x1D,0,0,midichannel},  // MidiData[3] = Pot[3]  : CC 0x1D = Undefined
  {0x1E,0,0,midichannel},  // MidiData[4] = Pot[4]  : CC 0x1E = Undefined
  {0x1F,0,0,midichannel},  // MidiData[5] = Pot[5]  : CC 0x1F = Undefined
  {0x59,0,0,midichannel},  // MidiData[6] = Pot[6]  : CC 0x59 = Undefined
  {0x5A,0,0,midichannel},  // MidiData[7] = Pot[7]  : CC 0x5A = Undefined
  // switches
  {0x66,0,0,midichannel},  // MidiData[8] = switches[0] = toggle switch 1 (with LED)  : CC 0x66 = Undefined
  {0x67,0,0,midichannel},  // MidiData[9] = switches[1] = toggle switch 2 (with LED)  : CC 0x67 = Undefined
  {0x68,0,0,midichannel},  // MidiData[10] = switches[2] = toggle switch 3 (with LED) : CC 0x68 = Undefined
  {0x69,0,0,midichannel},  // MidiData[11] = switches[3] = toggle switch 4 (with LED) : CC 0x69 = Undefined
  {0x6A,0,0,midichannel},  // MidiData[12] = switches[4] = toggle switch 5 (with LED) : CC 0x6A = Undefined
  {0x6B,0,0,midichannel},  // MidiData[13] = switches[5] = toggle switch 6 (with LED) : CC 0x6B = Undefined
  {0x6C,0,0,midichannel},  // MidiData[14] = switches[6] = toggle switch 7 (with LED) : CC 0x6C = Undefined
  {0x6D,0,0,midichannel},  // MidiData[15] = switches[7] = toggle switch 8 (with LED) : CC 0x6D = Undefined
  // Playback controls
  {0x50,0,0,midichannel},  // MidiData[16] = switches[8] = Playback switch 1 STOP (with LED)      : CC 0x50  General Purpose Controller 5
  {0x51,0,0,midichannel},  // MidiData[17] = switches[9] = Playback switch 2 PLAY (with LED)      : CC 0x51  General Purpose Controller 5
  {0x52,0,0,midichannel},  // MidiData[18] = switches[10] = Playback switch 3 REC (with LED)      : CC 0x52  General Purpose Controller 5
  {0x53,0,0,midichannel},  // MidiData[19] = switches[11] = Playback switch 4 SPEAKER (with LED)  : CC 0x53  General Purpose Controller 5
  // Navigation and control
  {0x10,0,0,midichannel},  // MidiData[20] = switches[12] = Navig switch 1 LEFT     : CC 0x10 = General Purpose Controller 1
  {0x11,0,0,midichannel},  // MidiData[21] = switches[13] = Navig switch 2 UP       : CC 0x11 = General Purpose Controller 2
  {0x12,0,0,midichannel},  // MidiData[22] = switches[14] = Navig switch 3 DOWN     : CC 0x12 = General Purpose Controller 3
  {0x13,0,0,midichannel},  // MidiData[23] = switches[15] = Navig switch 4 RIGHT    : CC 0x13 = General Purpose Controller 4
  {0x14,0,0,midichannel},  // MidiData[24] = switches[19] = Function switch Control : CC 0x14 = Undefined
  {0x15,0,0,midichannel},  // MidiData[25] = switches[16] = Function switch Sound   : CC 0x15 = Undefined
  {0x16,0,0,midichannel},  // MidiData[26] = switches[20] = Function switch F1      : CC 0x16 = Undefined
  {0x17,0,0,midichannel},  // MidiData[27] = switches[17] = Function switch F2      : CC 0x17 = Undefined
  {0x18,0,0,midichannel},  // MidiData[28] = switches[21] = Function switch Enter   : CC 0x18 = Undefined
  {0x19,0,0,midichannel},  // MidiData[29] = switches[18] = Function switch ESC     : CC 0x19 = Undefined
  // foot switches
  {0x04,0,0,midichannel},  // MidiData[30] = footswitch 1A    : CC 0x04 = Foot Controller
  {0x64,0,0,midichannel},  // MidiData[31] = footswitch 1B    : CC 0x64 = Damper Pedal on/off (Sustain)
  {0x43,0,0,midichannel},  // MidiData[32] = footswitch 2A    : CC 0x43 = Soft pedal On/Off
  {0x44,0,0,midichannel},  // MidiData[33] = footswitch 2B    : Cc 0x44 = Legato Foot Switch
  // pedals
  {0x01,0,0,midichannel},  // MidiData[34] = Pedal 1          : CC 0x01 = Modulation Wheel or Lever
  {0x0B,0,0,midichannel},  // MidiData[35] = Pedal 2          : Cc 0x0B = Expression Controller
  // volume
  {0x19,100,0,midichannel}   // MidiData[36] = Rotary Encoder : CC 0x07 = Volume control
};

typedef struct                  // define struct with all values related to each switch
{
  bool toggle;                  // Toggle function or momentary: TRUE = Toggle
  byte previousReadValue;       // Previous value read from footswitch, used for doggle function
  long debounceTimeStamp;       // Timestamp for debounce
  byte pin;                     // Input pin
} footSwitch;

footSwitch FootSwitchConfig[4] =  // Configuration for the footswitch inputs
{ // toggle enable, debounce timestamp, previous read value, Input pin
  {0,0,0, FS_1_A_PIN},
  {0,0,0, FS_1_B_PIN},
  {0,0,0, FS_2_A_PIN},
  {0,0,0, FS_2_B_PIN},
};

// OLED contructor
//U8G2_SSD1309_128X64_NONAME0_1_4W_HW_SPI u8g2(U8G2_R2,/* cs=*/ 2, /* dc=*/ A5, /* reset=*/ A4); // KORE Hardware
U8G2_SSD1309_128X64_NONAME0_1_4W_HW_SPI u8g2(U8G2_R2,/* cs=*/ 10, /* dc=*/ 9, /* reset=*/ 7);   // nRF52840-DK debug setup

// Adafruit BLE midi
BLEDis bledis;
BLEMidi blemidi;
MIDI_CREATE_BLE_INSTANCE(blemidi);  // Create a new instance of the Arduino MIDI Library, and attach BluefruitLE MIDI as the transport.

// Global variables
char pedal1status = false;                 // Status of the expression pedals
char pedal2status = false;                 // Status of the expression pedals
int direction = 1;

void setup()
{
  // Configure IO
  digitalWrite(PEDAL1_RING_PIN, HIGH);      // High drive for Pedal input (ring connection->High side of pedal pot)
  digitalWrite(PEDAL2_RING_PIN, HIGH);      // High drive for Pedal input (ring connection->High side of pedal pot)
  pinMode(PEDAL1_RING_PIN,INPUT_PULLUP);    // Set up expression pedal input with pullup, tip is pulled to GND through Jack switch
  pinMode(PEDAL2_RING_PIN,INPUT_PULLUP);    // Set up expression pedal input with pullup, tip is pulled to GND through Jack switch
  pinMode(FS_1_A_PIN,INPUT_PULLUP);         // Set up foot switch inputs with pullup
  pinMode(FS_1_B_PIN,INPUT_PULLUP);         // Set up foot switch inputs with pullup
  pinMode(FS_2_A_PIN,INPUT_PULLUP);         // Set up foot switch inputs with pullup
  pinMode(FS_2_B_PIN,INPUT_PULLUP);         // Set up foot switch inputs with pullup
  // Set up ADC (For expression pedal measurements
  analogReference(AR_VDD4);                 // Set the analog reference to VCC (VDD/4 with gain 4)
  analogReadResolution(ADC_RESOLUTION);     // Set ADc resolution
  // Set up Graphics, I2C and serial
  u8g2.begin();                             // Initialize graphics lib
  u8g2.setContrast(0xff);
  Wire.begin(8);                            // Setup and join i2c bus with address #8
  Wire.onReceive(receiveEvent);             // register event for I2C received data handling
  Serial.begin(115200);                     // start serial for debug output
  // Bluetooth Configuration Note: All config***() function must be called before begin()
  Bluefruit.configPrphBandwidth(BANDWIDTH_MAX);   // Config the peripheral connection with maximum bandwidth
  Bluefruit.begin();
  Bluefruit.setName("Bluefruit52 MIDI");
  Bluefruit.setTxPower(8);              // 8dBm transmit power
  Bluefruit.autoConnLed(true);          // Setup the on board blue LED to be enabled on CONNECT
  bledis.setManufacturer("Adafruit Industries"); // Configure and Start Device Information Service
  bledis.setModel("Bluefruit Feather52");
  bledis.begin();
  startAdv();                             // Set up and start advertising
  // Set up MIDI library
  MIDI.begin(MIDI_CHANNEL_OMNI);          // Initialize MIDI, and listen to all MIDI channels
  MIDI.setHandleNoteOn(handleNoteOn);     // Attach the handleNoteOn function to the MIDI Library
  MIDI.setHandleNoteOff(handleNoteOff);   // Do the same for MIDI Note Off messages.
  Scheduler.startLoop(midiRead);          // Start MIDI read loop
  // Set up rotary encoder HW module interrupt driven
  RotaryEncoder.begin(ROTARY_ENC_PINA, ROTARY_ENC_PINB);  // Initialize encoder
  RotaryEncoder.setCallback(encoder_cbk); // Set callback function
  RotaryEncoder.start();                  // Start encoder
}

void startAdv(void)
{
  Bluefruit.Advertising.addFlags(BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE);  // Set General Discoverable Mode flag
  Bluefruit.Advertising.addTxPower();              // Advertise TX Power
  Bluefruit.Advertising.addService(blemidi);       // Advertise BLE MIDI Service
  Bluefruit.ScanResponse.addName();                // Secondary Scan Response packet (optional) since there is no room for 'Name' in Advertising packet
  Bluefruit.Advertising.restartOnDisconnect(true); //Enable auto advertising if disconnected
  Bluefruit.Advertising.setInterval(32, 244);      // Interval:  fast mode = 20 ms, slow mode = 152.5 ms, in unit of 0.625 ms
  Bluefruit.Advertising.setFastTimeout(30);        // Timeout (seconds) for fast mode is 30 seconds
  Bluefruit.Advertising.start(0);                  // Start(timeout) with timeout = 0 will advertise forever (until connected)
}

void encoder_cbk(int step)                         // Callback function for Rotary encoder
{
  Serial.print("Rotary Encoder :");
  if ( step > 0 )                                  // if rotation towards right
  {
    if (MidiData[36].value < 127)                  // If not yet at maximum
    {
      MidiData[36].value = MidiData[36].value + 2; // Increase volume
    }
    Serial.print("Right ");
    Serial.println(step);
  }
  else                                             // if rotation towards right
  {
    if (MidiData[36].value > 0)                    // If not yet at minimum
    {
      MidiData[36].value = MidiData[36].value - 2; // Decrease  volume
    }
    Serial.print("Left ");
    Serial.println(step);
  }
}

void handleNoteOn(byte channel, byte pitch, byte velocity)
{
  Serial.printf("Note on: channel = %d, pitch = %d, velocity - %d", channel, pitch, velocity);    // Log when a note is pressed.
  Serial.println();
}

void handleNoteOff(byte channel, byte pitch, byte velocity)
{
  Serial.printf("Note off: channel = %d, pitch = %d, velocity - %d", channel, pitch, velocity);  // Log when a note is released.
  Serial.println();
}

void midiRead()
{

  if (! Bluefruit.connected()) {    // Don't continue if we aren't connected.
    return;
  }
  if (! blemidi.notifyEnabled()) {  // Don't continue if the connected device isn't ready to receive messages.
    return;
  }
  MIDI.read();                      // read any new MIDI messages
}

// function that executes whenever data is received from I2C master, registered as an event in setup()
void receiveEvent(int howMany)
{
  for (int i=0; i<30; i++)          //Receive 30 bytes from slave board (button and pot encoder)
  {
    byte c = Wire.read();           // receive byte as a integer
    MidiData[i].value = c;
  }
  if (Wire.available())
  {
      Serial.print("Error more than 30 char received");
  }
}

// Function to read foot switch inputs
void CheckFS(void)
{
  for (int i=0; i<4; i++)                                     //Check all 4 foot switches
  {
    byte readValue = !digitalRead (FootSwitchConfig[i].pin);  // Read current state of the pin/Footswitch inverted (active low)
    if (FootSwitchConfig[i].toggle == true)                   // if switch is toggle type
    { //  Current read FS input  &&  Previos FS value ......        && longer than debounce time since last change
      if ((readValue == true) && (FootSwitchConfig[i].previousReadValue == false) && ((millis() - FootSwitchConfig[i].debounceTimeStamp) > DEBOUNCE)) //If switch is active, last time was inactive, and more than "DEBOUNCE" ms since last toggle
      {
        if (MidiData[i+30].value > 0)                         // If current walue is on...
        {
          MidiData[i+30].value = 0;                           // ..set foot switch value to off (0)
        }
        else                                                  // otherwise, current value is off..
        {
          MidiData[i+30].value = 127;                         // and new value is set to to on
        }
        FootSwitchConfig[i].debounceTimeStamp = millis();     // Reset timestamp for when toggle happened (for debouncing logic)
      }
    }
    else                                                      //if switch is not toggle type, it should be momentary
    {
      if ((readValue != FootSwitchConfig[i].previousReadValue) && ((millis() - FootSwitchConfig[i].debounceTimeStamp) > DEBOUNCE)) //If button has changed state and more than "DEBOUNCE" ms since last press
      {
        MidiData[i+30].value = readValue*127;                 // Set value to currently read value (0=0, 1=127)
        FootSwitchConfig[i].debounceTimeStamp = millis();     // Reset timestamp for when toggle happened (for debouncing logic)
      }
    }
    FootSwitchConfig[i].previousReadValue = readValue;        // Remember recent read value until next check
  }
}

void CheckPedal (void)
{
/*  int pedal1value = analogRead(PEDAL1_TIP_PIN);     //Measure voltage on pedal tip connection
  int pedal2value = analogRead(PEDAL2_TIP_PIN);     //Measure voltage on pedal tip connection
  int pedal1ring = analogRead(PEDAL1_RING_PIN);   //Measure voltage on pedal ring connection
  int pedal2ring = analogRead(PEDAL2_RING_PIN);   //Measure voltage on pedal ring connection
  Serial.print("1t: ");
  Serial.print(pedal1value);
  Serial.print(" 1r: ");
  Serial.print(pedal1ring);
  Serial.print(" 2t: ");
  Serial.print(pedal2value);
  Serial.print(" 2r: ");
  Serial.println(pedal2ring);
  Serial.print(" ADCMAX ");*/
  if (analogRead(PEDAL1_RING_PIN) > (ADC_MAX_VALUE - 15)) // If pedal1 ring connection is pulled to VCC -> no pedal
  {
//      Serial.println("Pedal1 not connected");
      pedal1status = PEDAL_NOT_CONNECTED;
  }
  else if (analogRead(PEDAL1_RING_PIN) < 10 )             // Check if pedal1 ring connection is pulled to ground -> error
  {
//    Serial.println("Pedal1 shorted to gnd, make sure to fully insert Pedal");
    pedal1status = PEDAL_ERROR;
  }
  else
  {
    pinMode(PEDAL1_RING_PIN,OUTPUT);                      // output VCC on pedal pot
    int pedal1value = analogRead(PEDAL1_TIP_PIN);         // Measure voltage on pedal tip connection again, now with VCC driven and not just pullup
    pinMode(PEDAL1_RING_PIN,INPUT_PULLUP);                // Turn off VCC drive and tristate with pullup
    pedal1status = PEDAL_CONNECTED;
//    Serial.print("Pedal1 value");
//    Serial.println(pedal1value);
    pedal1value = pedal1value/ADC_RESULT_DIVIDER;         // Divide down ADC result to be in MIDI range (0-127)
    if ((pedal1value > MidiData[34].value) || (pedal1value < MidiData[34].value - 1) || (pedal1value == 0))  // add hysteresis before moving down
    {
      MidiData[34].value = pedal1value;                   // Update MIDI value
    }
  }

  if (analogRead(PEDAL2_RING_PIN) > (ADC_MAX_VALUE - 15)) // If pedal1 ring connection is pulled to VCC -> no pedal
  {
      pedal2status = PEDAL_NOT_CONNECTED;
  }
  else if (analogRead(PEDAL2_RING_PIN) < 10 )             // Check if pedal1 ring connection is pulled to ground -> error
  {
    Serial.println("Pedal2 Ring connection shorted to gnd, make sure to fully insert an expression pedal");
    pedal2status = PEDAL_ERROR;
  }
  else
  {
    pinMode(PEDAL2_RING_PIN,OUTPUT);                      // output VCC on pedal pot
    int pedal2value = analogRead(PEDAL2_TIP_PIN);         // Measure voltage on pedal tip connection again, now with VCC driven and not just pullup
    pinMode(PEDAL2_RING_PIN,INPUT_PULLUP);                // Turn off VCC drive and tristate with pullup
    pedal2status = PEDAL_CONNECTED;
//    Serial.print("Pedal2 value");
//    Serial.println(pedal2value);
    pedal2value = pedal2value/ADC_RESULT_DIVIDER;         // Divide down ADC result to be in MIDI range (0-127)
    if ((pedal2value > MidiData[35].value) || (pedal2value < MidiData[35].value - 1) || (pedal2value == 0))  // Only change value if higher, 2 lower or zero to avoid flickering.
    {
      MidiData[35].value = pedal2value;                   // Update MIDI value
    }
  }
}

void DrawMainScreen (void)
{
  u8g2.firstPage();
  do
  {
    // Build frame
    u8g2.drawFrame(1,0,127,63);         // Draw KORE window outline frame for reference while debugging (x,y,width, height)
    u8g2.drawLine(1, 12, 128, 12);      //Horisontal divider
    u8g2.drawLine(54, 1, 54, 11);       //Vertical divider
    u8g2.drawLine(1, 50, 112, 50);      //Horisontal divider
    u8g2.drawLine(74, 50, 74, 62);      //Vertical divider between connection status and channel
    u8g2.drawLine(112,12,112,62);       //Vertical divider for volume slider box

    // Draw Foot switch status icons
    u8g2.setFont(u8g2_font_6x12_t_symbols);
    for (int i=0; i<4;i++)
    {
      if (MidiData[30+i].value)                 // If footswitch active draw closed square
      {
        u8g2.drawGlyph((18+(i*9)), 10, 0x25ac); // hex 12fa closed square. Draw at pos 18 with 9pix spacing
      }
      else                                      // If footswitch inactive draw open square
      {
        u8g2.drawGlyph((18+(i*9)), 10, 0x25ad); // hex 25ef open square
      }
    }
    // Draw Footswitch text and pedal values
    u8g2.setFont(u8g2_font_finderskeepers_tf);
    u8g2.drawStr(4,10,"FS:");
    if ((pedal1status == PEDAL_ERROR) || (pedal2status == PEDAL_ERROR))
    {
      u8g2.drawStr(66,10,"Pedal error");
    }
    else
    {
      u8g2.drawStr(58,10,"P1:");
      if (pedal1status == PEDAL_CONNECTED)
      {
        u8g2.setCursor(70, 10);
        u8g2.print(u8x8_u8toa(MidiData[34].value ,3));
      }
      else                                          // Pedal not connected
      {
        u8g2.drawStr(70,10,"---");
      }

      u8g2.drawStr(92,10,"P2:");
      if (pedal2status == PEDAL_CONNECTED)
      {
        u8g2.setCursor(107, 10);
        u8g2.print(u8x8_u8toa(MidiData[35].value ,3));
      }
      else                                          // Pedal not connected
      {
        u8g2.drawStr(107,10,"---");
      }
    }
    // Print Midi channel
    u8g2.setCursor(81,60);
    u8g2.print("ch: ");
    u8g2.print(u8x8_u8toa(midichannel, 2));
    // Print connection status
    if (Bluefruit.connected())
    {
      u8g2.drawStr(16,60,"connected");
    }
    else
    {
      u8g2.drawStr(9,60,"disconnected");
    }
    //Draw pot values
    u8g2.setFont(u8g2_font_helvR10_tf);
    u8g2.setCursor(3, 28);
    u8g2.print(u8x8_u8toa(MidiData[0].value ,3));
    u8g2.setCursor(31,28);
    u8g2.print(u8x8_u8toa(MidiData[1].value ,3));
    u8g2.setCursor(59,28);
    u8g2.print(u8x8_u8toa(MidiData[2].value, 3));
    u8g2.setCursor(87,28);
    u8g2.print(u8x8_u8toa(MidiData[3].value, 3));
    u8g2.setCursor(3,46);
    u8g2.print(u8x8_u8toa(MidiData[4].value, 3));
    u8g2.setCursor(31,46);
    u8g2.print(u8x8_u8toa(MidiData[5].value, 3));
    u8g2.setCursor(59,46);
    u8g2.print(u8x8_u8toa(MidiData[6].value, 3));
    u8g2.setCursor(87,46);
    u8g2.print(u8x8_u8toa(MidiData[7].value, 3));
    //Draw Volume slider
    u8g2.drawBox(119,18,2,38);                            // Draw slider  38 pixels long, 2 pixels wide
    char sliderPosition = ((37*MidiData[36].value)/127);  // Slider knob position. Slider lenght - 1 (37) used to look more natural
    u8g2.drawFrame (116,53-sliderPosition,8,5);           // Draw slider knob
    u8g2.setDrawColor(0);                                 // Change to erase
    u8g2.drawBox (117,54-sliderPosition,6,3);             // Erase slider pathway from knob
    u8g2.setDrawColor(1);
  }
  while ( u8g2.nextPage() );
}

void loop()
{
  CheckFS();                                                  // Check state of Foot Switches
  CheckPedal();                                               // Check state of Pedals
  //  Check if any controller have changed value TODO: change pot/button code to send midi data directly and use one function
  for (int i=0; i<37; i++)                                    // Run through all MIDI values and send CC data if values have changed
  {
    if (MidiData[i].value != (MidiData[i].previousSentValue)) // If data has changed
    {
      MIDI.sendControlChange (MidiData[i].cccode, MidiData[i].value, MidiData[i].channel);
      MidiData[i].previousSentValue = MidiData[i].value;
    }
  }
  DrawMainScreen();
}
